/*
 * Soubor:  crcal.c
 * Datum:   17.03.2011
 * Autor:   Marek Nožka, nozka <@t> spseol <d.t> cz
 * Licence: GNU/GPL 
 * Úloha:   Komplexní kalkulátor s reverzní polskou notací
 ****************************************************/
#define _ISOC99_SOURCE
#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdbool.h>
#include <math.h>
#include <complex.h>

/* ************************************************ */

/* ***************    Makra    ******************** */
#define CMDLEN 128              // počet znaků jednoho příkazu
#define MSGLEN 1024             // počet znaků při tisku obsahu zásobníku
#define STACKC 64               // velikost zásobníku

/* *********** Prototypy  funkcí ****************** */

#include "crcal.h"

/* *********** Globální proměnné ****************** */
double (*torad) (double) = deg2rad;     // Odkaz na funkci, která určuje, ...
double (*radto) (double) = rad2deg;     // zda se používají radiány nebo stupně

void (*f_print_s) (char *) = stackEprnt;        // Odkaz na funkci, která určuje, ...
                                        // tvar komplexního čísla


/****************************************************
 *               Hlavní program.
 ****************************************************/
int main(void)
{
    double rnumber;             // číslo se kterým se má počítat
    double inumber;             // číslo se kterým se má počítat
    char enter[CMDLEN];         // vstup od uživatele
/*    char message[MSGLEN];*/
/*    double variables[VARIABLEC] = { 0 };*/

    /* Čte řádky až do konce souboru */
    //while (scanf("%s", enter) != EOF) {
    printStack();
    while (readEnter(enter) != EOF) {
        if (sscanf(enter, "%lg,%lg", &rnumber, &inumber) == 2) {
            push(rnumber + inumber * I);
        } else if (sscanf(enter, "%lgL%lg", &rnumber, &inumber) == 2) {
            push(rnumber * cos(torad(inumber)) +
                 I * rnumber * sin(torad(inumber)));
        } else if (sscanf(enter, ",%lg", &inumber) == 1) {
            push(I * inumber);
        } else if (sscanf(enter, "%lg", &rnumber) == 1) {
            push(rnumber);
        } else if (enter[0] == '+') {
            makeWrap(makePlus, enter[0]);
        } else if (enter[0] == '-') {
            makeWrap(makeMinus, enter[0]);
        } else if (enter[0] == '*') {
            makeWrap(makeTimes, enter[0]);
        } else if (enter[0] == '/') {
            makeWrap(makeDivid, enter[0]);
        } else if (strcmp(enter, "deg") == 0 || strcmp(enter, "d") == 0) {
            setdeg();
        } else if (strcmp(enter, "rad") == 0 || strcmp(enter, "r") == 0) {
            setrad();
        } else if (strcmp(enter, "rem") == 0 || strcmp(enter, "m") == 0) {
            pop();
            printStack();
        } else if (strcmp(enter, "clr") == 0 || strcmp(enter, "c") == 0) {
            stackclear();       // vymaže celý zásobník
        } else if (strcmp(enter, "print") == 0 || strcmp(enter, "p") == 0) {
            printStack();       // jen tiskne zásobník
        } else if (strcmp(enter, "alg") == 0 || strcmp(enter, "a") == 0) {
            f_print_s = stackAprnt;     // přepne na algebraický tvar
        } else if (strcmp(enter, "exp") == 0 || strcmp(enter, "e") == 0) {
            f_print_s = stackEprnt;     // Přepne na exponenciální tvar
        } else if (strcmp(enter, "help") == 0 || strcmp(enter, "h") == 0) {
            printHelp();
            printStack();
        } else if (strcmp(enter, "quit") == 0 || strcmp(enter, "q") == 0) {
            return 0;
        } else {
            fprintf(stderr, " error > %s\n", enter);
            fprintf(stderr, " error > Neplatný vstup\n");
            printStack();
        }
    }

    stackclear();
    return 0;
}

/****************************************************
 *                 Funkce
 ****************************************************/

int readEnter(char enter[])     //%.h
{
    int c, ret;
    while (((c = getchar()) == ' ') || c == '\t' || c == '\n');
    if (c == EOF) {
        return EOF;
    }
    ungetc(c, stdin);

    ret = scanf("%s", enter);
/*    printf("#### ctu #%s#\n",enter);*/

    while ((c = getchar()) == ' ' || c == '\t' );
    if (c == '\n') {
        printf(" > ");
    } else {
        ungetc(c, stdin);
    }

    return ret;
}

void printStack()               //%.h
{
    char message[MSGLEN];
    f_print_s(message);         // zjistím, co je v zásobníku
    printf(">>> %s\n > ", message);        // ... a vytisknu to 
}

/* Tiskne malou nápovědu */
void printHelp(void)            //%.h
{
    printf
        ("Kalkulátor s polskou reverzní notací. Pořadí matematických\n"
         "operací je dáno přímo zápisem. Například: \n\n"
         "(5 * (4 + 3) ) / (10-3) odpovídá 5 4 3 + * 10 3 - /\n"
         "\nKalkulátor počítá v komplexních číslech. "
         "Podporované matematické operace:\n"
         "+ - * /\n\n"
         "deg|d    přepne pro počítání ve stupních\n"
         "rad|r    přepne pro počítání v radiánech\n"
         "rem|m    vymaže poslední číslo ze zásobníku\n"
         "alg|a    print vypisuje v algebraickém tvaru\n"
         "exp|e    print vypisuje v exponenciálním tvaru\n"
         "clr|c    vymaže celý zásobník\n"
         "quit|q   ukončí program\n" "\n");
}

/****************************************************
 *            Proměnné zásobníku
 ****************************************************/

complex double stack[STACKC];   // zásobník
int stackc;                     // pozice v zásobníku

/* Vloží do zásobníku */
int push(double complex number) //%.h
{
    if (stackc < STACKC) {
        stack[stackc] = number;
        return ++stackc;
    } else
        return EOF;
}

/* Vyjme ze zásobníku */
complex double pop()            //%.h
{
    if (stackc > 0)
        return stack[--stackc];
    else
        return NAN;
}

/* Vymaže celý zásobník */
void stackclear(void)           //%.h
{
    stackc = 0;
}

/* Funkce do řetězce s vytiskne obsah zásobníku */
void stackAprnt(char s[])       //%.h
{
    char bagr[CMDLEN];

    strcpy(s, "[ ");
    for (int i = 0; i < stackc; i++) {
        strcat(s, "(");
        complex double number = stack[i];
        /* tisk reálné části */
        if (creal(number) != 0) {
            sprintf(bagr, "%.4g", creal(number));
            strcat(s, bagr);
        }
        /* tisk imaginární části */
        if (cimag(number) != 0) {
            strcat(s, cimag(number) > 0 ? "+j" : "-j");
            sprintf(bagr, "%.4g",
                    cimag(number) > 0 ? cimag(number) : -cimag(number));
            strcat(s, bagr);
        }
        strcat(s, ") ");
    }
    strcat(s, "]");
}

/* Funkce do řetězce s vytiskne obsah zásobníku */
void stackEprnt(char s[])       //%.h
{
    char bagr[CMDLEN];

    strcpy(s, "[ ");
    for (int i = 0; i < stackc; i++) {
        strcat(s, "(");
        complex double number = stack[i];
        sprintf(bagr, "%.4gL%.4g", cabs(number), radto(carg(number)));
        strcat(s, bagr);
        if (radto == rad2deg) {
            strcat(s, "°");
        }
        strcat(s, ") ");
    }
    strcat(s, "]");
}

/* Funkce převádí stupně na radiány */
double deg2rad(double deg)      //%.h
{
    return M_PI * deg / 180;
}

double rad2rad(double rad)      //%.h
{
    return rad;
}

double rad2deg(double deg)      //%.h
{
    return 180 * deg / M_PI;
}

/*    plus      */
double complex makePlus(double complex x, double complex y)     //%.h
{
    return x + y;
}

/*    minus      */
double complex makeMinus(double complex x, double complex y)    //%.h
{
    return x - y;
}

/*    krát      */
double complex makeTimes(double complex x, double complex y)    //%.h
{
    return x * y;
}

/*    děleno      */
double complex makeDivid(double complex x, double complex y)    //%.h
{
    return x / y;
}

/* Funkce vyzvedne ze zásobníku 2 čísla, provede s nimi
 * matematickou operace danou parametrem function a výsledek
 * uloží zpět do zásobníku.
 * Funkce vrací FALSE pokud operaci nelze provést.
 */
bool makeOper(operation function)       //%.h
{
    complex double x, y;
    if (isnan(y = pop())) {
        return false;
    }
    if (isnan(x = pop())) {
        push(y);
        return false;
    }
    push(function(x, y));
    return true;
}


/* Tato funkce jen vhodně obaluje make2param a komunikuje 
 * s uživatelem*/
void makeWrap(operation function, char operationName)   //%.h
{
    if (!makeOper(function)) {
        fprintf(stderr, " error > %c vyžaduje dva parametry\n",
                operationName);
    }
    printStack();
}


/* Funkce přepne do módu radiánů nebo stupňů*/
void setrad(void)               //%.h
{
    torad = rad2rad;
    radto = rad2rad;
}

void setdeg(void)               //%.h
{
    torad = deg2rad;
    radto = rad2deg;
}
